#! /usr/bin/python

"""
The Petmail agent runs on the local system, holds the private key, and acts
as an SMTP/POP/IMAP proxy for a local legacy MUA.
"""

import os.path

from twisted.application import service
from twisted.internet import defer, reactor
from twisted.python.components import implements
from twisted.python import log

from petmail import addressbook, owner, message, permission
from petmail import interfaces, processors, crypto, util, request
import petmail.log

class Agent(service.MultiService):
    """This is the main object which defines the Petmail agent. There will
    be exactly one of these in each agent process (it can accomodate
    multiple senders). This is the object which is serialized to record the
    persistent state of the service, including the contents of the
    addressbook and all the stored Permissions.

    """

    debugProcessing = False

    def __init__(self, basedir):
        service.MultiService.__init__(self)
        self.basedir = basedir
        self.senders = {}
        self.addressbooks = {}
        self.inboundProcessors = []
        self.outboundProcessors = []
        self.addressLookupAgents = []
        self.addressPublishAgents = []
        self.smtpScheme = "Mail:localhost:petmail-user@localhost"
        self.nntpScheme = None

        self.nextRequestID = 1
        self.outstandingRequests = {}
        self.scheduledEvents = [] # this is a list of (when, Request) tuples
        self.scheduledEventTimer = None

        # these lists track messages that are being processed. Each one must
        # support a .why() method which returns a string to indicate what the
        # message is waiting for. They should also have a timeout mechanism.
        self.pendingInboundMessages = []
        self.pendingOutboundMessages = []

        self.eventLog = petmail.log.Log("agent")
        if 0:
            ds = debug.DebugService()
            ds.setServiceParent(self)
        self.defaultReturnPermission = permission.OwnerPermission()
        self.log("Agent initialized")

    def log(self, *args):
        self.eventLog.add(*args)

    def init(self):
        self.populateProcessors()
        self.crypto = crypto.EncryptionAgent(os.path.join(self.basedir,
                                                          "crypto"))

    def __getstate__(self):
        timer = self.scheduledEventTimer
        state = service.MultiService.__getstate__(self)
        state['scheduledEventTimer'] = None
        assert self.scheduledEventTimer == timer
        return state

    def startService(self):
        self.log("Agent.startService")
        service.MultiService.startService(self)
        reactor.callLater(6, self.checkScheduledEvents)

    def stopService(self):
        self.log("Agent.stopService")
        if self.scheduledEventTimer:
            self.scheduledEventTimer.cancel()
            self.scheduledEventTimer = None
        return service.MultiService.stopService(self)

    def createIdentity(self):
        """Returns a Deferred which will fire with the new OwnerEntry.
        """
        me = owner.OwnerEntry()
        me.setGenericPermission(permission.OwnerPermission())

        d = self.crypto.generateKey("name")
        d.addCallback(self._createIdentityDone, me)
        return d
        
    def _createIdentityDone(self, key, me):
        me.setKey(key)
        return me

    def addOwner(self, owner):
        self.senders[owner.ID] = owner
        self.addressbooks[owner] = addressbook.AddressBook(owner)
        owner.setServiceParent(self)
    def removeOwner(self, owner):
        owner.disownServiceParent()
        del self.addressbooks[owner]
        del self.senders[owner.ID]
    def getOwners(self):
        ownerlist = self.addressbooks.keys()
        ownerlist.sort(lambda x,y: cmp(x.getName(), y.getName()))
        return ownerlist
    def getOwnerByName(self, name):
        for sender in self.senders.values():
            if name == sender.getName():
                return sender
        raise KeyError("no owner named '%s'" % name)


    def setSMTP(self, scheme):
        self.smtpScheme = scheme

    def setNNTP(self, scheme):
        self.nntpScheme = scheme

    def setAddressLookup(self, scheme):
        while self.addressLookupAgents:
            self.addressLookupAgents.pop()
        while self.addressPublishAgents:
            self.addressPublishAgents.pop()
        if not len(scheme):
            return
        for s in scheme.split(","):
            la = addressbook.getLookupAgentFromString(s)
            self.registerAddressLookupAgent(la)
    def getAddressLookup(self):
        return ",".join([la.describe() for la in self.addressLookupAgents])

    def registerAddressLookupAgent(self, agent):
        self.addressLookupAgents.append(agent)
        self.addressPublishAgents.append(agent)
    def unregisterAddressLookupAgent(self, agent):
        self.addressLookupAgents.remove(agent)
        self.addressPublishAgents.remove(agent)


    def getAddressBook(self, senderID):
        return self.addressbooks[self.getSender(senderID)]
    def getSender(self, senderID):
        return self.senders[senderID]

    def retrieveAddress(self, address):
        for method in self.addressLookupAgents:
            d = method.lookup(self, address)
            if isinstance(d, defer.Deferred):
                return d
            if d:
                return defer.succeed(d)
        else:
            # TODO: lookup always returns a Deferred. This clause needs to
            # be in an errback.
            raise interfaces.Unsendable, \
                  "none of my lookup agents could handle address '%s'" % \
                  address


    def populateProcessors(self):
        self.inboundProcessors = [pc() for pc in processors.allInbound]
        self.outboundProcessors = [pc() for pc in processors.allOutbound]

    def getInboundProcessors(self):
        return self.inboundProcessors[:]

    def getOutboundProcessors(self):
        return self.outboundProcessors[:]

    def getTicketAgents(self):
        # override this to provide ticket agents
        return []

    def addressBookChanged(self, addrbook):
        # override this to update something when the address book changes
        pass

    # sending

    def sendMessage(self, message):
        self.pendingOutboundMessages.append(message)
        d = message.startProcessing()
        d.addCallback(self.deliverMessage)
        d.addErrback(self.sendMessageFailed, message)
        return d

    def deliverMessage(self, message):
        if message.worthSending:
            message.deliver()
        message.dump()
        self.pendingOutboundMessages.remove(message) # TODO: only upon ack
        return True # finished sending.. TODO: return status?

    def sendMessageFailed(self, why, message):
        print "sendMessageFailed", why
        self.pendingOutboundMessages.remove(message)

    # things that may happen during send

    def grantReturnPermission(self, entry, message):
        # add Specific Permission to this address book entry
        p = self.defaultReturnPermission.copy()
        entry.setTheirPermission(p)
        entry.modified()


    # receiving

    def receiveMessage(self, encrypted, debug=False):
        """Feed an inbound message to the decryption/verification stage. If
        accepted, the message will be put on the inbox and
        notifyNewMessage() will be called.

        This will be called by a transport layer, triggered by a mailbox
        poll or SMTP receive from the outside world.

        Returns a Deferred which will fire when the message has been fully
        processed. Remember that it might set up timers to make something
        happen later.
        
        @type encrypted: C{string}
        @param encrypted: the encrypted message blob
        """

        log.msg("Agent.receiveMessage")
        rm = message.ReceivedMessage(self, encrypted)
        if debug: rm.debug = True
        d = rm.decrypt()
        d.addCallback(self._receiveMessage2, rm)
        d.addErrback(self._receiveMessageFailed, rm)
        return d

    def _receiveMessage2(self, recipientEntry, rm):
        self.pendingInboundMessages.append(rm)
        d = rm.startProcessing()
        d.addCallback(self._receiveMessageDone)
        return d

    def _receiveMessageDone(self, msg):
        """This is called when an inbound message has been fully processed.
        Everything interesting has already been done by the Processors, so
        all that remains is to log receipt (and possibly send back an
        acknowledgement)."""
        self.pendingInboundMessages.remove(msg)
        msg.dump()

    def _receiveMessageFailed(self, why, msg):
        # an error occurred during processing, should log
        print "receiveMessageFailed", why
        self.pendingInboundMessages.remove(msg)
        msg.dump()

    def acceptRFC822(self, rm, why):
        # override this to do something with incoming messages
        assert implements(rm, interfaces.IReceivedMessage)

    def rejectRFC822(self, rm, why):
        # spam path, so this will only go into a log file, not bother the user
        print "rejectRFC822", why


    # request/response handling

    def getRequestID(self):
        reqid = self.nextRequestID
        self.nextRequestID += 1
        return reqid

    def addRequest(self, req):
        assert implements(req, interfaces.IRequest)
        assert not self.outstandingRequests.has_key(req.reqid)

    # scheduled events (restart-survivable DelayedCalls)

    def scheduleEvent(self, seconds, description, results):
        """I schedule an event to happen some time in the future. I return a
        Deferred which will fire with the given results when the time is
        reached."""

        req = request.Request(description, results)
        e = (util.now() + seconds, req)
        self.scheduledEvents.append(e)
        self.checkScheduledEvents()
        return req.deferred

    def checkScheduledEvents(self):
        self.log("Agent.checkScheduledEvents [%d]" % \
                 len(self.scheduledEvents))
        # remember that this is reentrant
        if self.scheduledEventTimer:
            if self.scheduledEventTimer.active():
                self.scheduledEventTimer.cancel()
            self.scheduledEventTimer = None
        now = util.now()
        ready = []
        timeout = None
        for ev in self.scheduledEvents[:]:
            when, req = ev
            if when <= now:
                self.log(" event ready to fire:", req.why())
                self.scheduledEvents.remove(ev)
                ready.append(req)
            else:
                self.log(" still waiting %d:" % (when-now), req.why())
                if timeout == None:
                    timeout = when
                else:
                    timeout = min(timeout, when)
        if timeout != None:
            self.log(" next event is in %d seconds" % (timeout-now))
            relative = (timeout+1) - now # wake up a trifle late
            timer = reactor.callLater(relative, self.checkScheduledEvents)
            self.scheduledEventTimer = timer
        for req in ready:
            req.fire() # reentrancy can happen here
        self.log(" Agent.checkScheduledEvents done")

    def grantInboundPermission(self, ownerID, recipientID, message):
        self.log("Agent.grantInboundPermission")
        owner = self.senders.get(ownerID)
        if not owner:
            return
        recipient = self.addressbooks[owner].lookupID(recipientID)
        if not recipient:
            return
        owner.grantPermissionTo(recipient, message)

    def refreshRecipient(self, ownerID, recipientID):
        owner = self.senders.get(ownerID)
        if not owner:
            return
        recipient = self.addressbooks[owner].lookupID(recipientID)
        if not recipient:
            return
        m = message.Message(self)
        m.setSenderEntry(owner)
        m.setRecipientEntry(recipient)
        self.sendMessage(m)

    # startup utility methods
    def run(self):
        self.startService()
        reactor.run()
        self.stopService()

    def stop(self):
        reactor.stop()

def run():
    a = Agent()
    a.startService()
    reactor.run()
    a.stopService()

if __name__ == '__main__':
    run()
